/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_FILLER_H_
#define AIDGE_CORE_FILLER_H_

#include <memory>
#include <random>  // normal_distribution, uniform_real_distribution

#include "aidge/data/Tensor.hpp"

namespace Aidge {

inline void calculateFanInFanOut(std::shared_ptr<Tensor> tensor,
                                 unsigned int& fanIn, unsigned int& fanOut) {
    AIDGE_ASSERT(
        tensor->nbDims() == 4,
        "Tensor need to have 4 dimensions to compute FanIn and FanOut.");
    // Warning: This function suppose NCXX data layout.
    // Aidge currently only support NCHW but this maybe not be true in the
    // future.
    DimSize_t batchSize = tensor->dims()[0];
    DimSize_t channelSize = tensor->dims()[1];
    AIDGE_ASSERT(batchSize != 0,
                 "Cannot calculate FanIn if tensor batch size is 0.");
    AIDGE_ASSERT(channelSize != 0,
                 "Cannot calculate FanOut if tensor channel size is 0.");
    fanIn = static_cast<unsigned int>(tensor->size() / batchSize);
    fanOut = static_cast<unsigned int>(tensor->size() / channelSize);
}
enum VarianceNorm { FanIn, Average, FanOut };

template <typename T>
void constantFiller(std::shared_ptr<Tensor> tensor, T constantValue);

template <typename T>
void normalFiller(std::shared_ptr<Tensor> tensor, double mean = 0.0,
                  double stdDev = 1.0);

template <typename T>
void uniformFiller(std::shared_ptr<Tensor> tensor, T min, T max);

template <typename T>
void xavierUniformFiller(std::shared_ptr<Tensor> tensor, T scaling = 1.0,
                         VarianceNorm varianceNorm = FanIn);
template <typename T>
void xavierNormalFiller(std::shared_ptr<Tensor> tensor, T scaling = 1.0,
                        VarianceNorm varianceNorm = FanIn);

template <typename T>
void heFiller(std::shared_ptr<Tensor> tensor, VarianceNorm varianceNorm = FanIn,
              T meanNorm = 0.0, T scaling = 1.0);
}  // namespace Aidge

#endif /* AIDGE_CORE_FILLER_H_ */
